@@ -15,7 +15,7 @@ use Carp '&croak';
 use DynaLoader ();
 use Exporter ();
  
-$VERSION = '0.004';
+$VERSION = '0.005';
 @ISA = qw( Exporter DynaLoader IO::File );
 
 @EXPORT = qw( &KEY_FIRST &KEY_LAST &KEY_EQ &KEY_EQ_BWD 
@@ -34,14 +34,16 @@ $VERSION = '0.004';
                                     &KEY_FIRST &KEY_LAST &KEY_EQ &KEY_EQ_BWD
                                     &KEY_GE &RBA_EQ &RBA_EQ_BWD
                                   ) ],
-                 FUNCTIONS => [ qw( &flush &forward
+                 FUNCTIONS => [ qw( 
+                                    &dynalloc &dynfree 
+                                    &flush &forward
                                     &getname &get_dcb &mvsopen &mvswrite 
                                     &pds_mem &remove &rewind &resetpos
                                     &sysdsnr &tmpnam 
                                     &vsamdelrec &vsamlocate &vsamupdate
                                     ) ], 
                  EXPERIMENTAL => [ qw( 
-                                    &dsname_level &dynalloc &dynfree 
+                                    &dsname_level 
                                     &svc99 &vol_ser 
                                     ) ], 
                );
@@ -79,13 +81,14 @@ OS390::Stdio - S/390 standard I/O functions via POSIX/XPG extensions
 
 =head1 SYNOPSIS
 
-    use OS390::Stdio qw( &get_dcb &getname &pds_mem &sysdsnr
+    use OS390::Stdio qw( &dynalloc &dynfree 
+                         &get_dcb &getname &pds_mem &sysdsnr
                          &mvsopen &mvswrite 
                          &flush &forward &rewind &resetpos
                          &remove &tmpnam 
                          &vsamdelrec &vsamlocate &vsamupdate
       # future dslist        &dsname_level &vol_ser 
-      # future SVC 99        &dynalloc &dynfree &svc99 
+      # future SVC 99        &svc99 
                        );
 
     @dslist = dsname_level("FRED");
@@ -102,6 +105,7 @@ OS390::Stdio - S/390 standard I/O functions via POSIX/XPG extensions
     close($fh);
     remove("dd:MYDD(MEM)");
     @members = pds_mem("//'SYS1.PARMLIB'");
+    @aliases = pds_mem("//'SYS1.PARMLIB'",1);
 
 =head1 DESCRIPTION
 
@@ -110,7 +114,7 @@ C stdio operations not available through Perl's CORE I/O functions.
 The specific routines are described below.  These functions are
 prototyped as unary operators, with the exception of C<mvsopen>
 which takes two arguments, C<mvswrite> which takes three arguments,
-C<dynalloc>, C<svc99> which take several arguments, and C<tmpnam> 
+C<svc99> which take several arguments, and C<tmpnam> 
 which takes none.
 
 All of the routines are available for export, though none are
@@ -142,13 +146,97 @@ actually call a method that OS390::Stdio doesn't provide.  This
 is done to save startup time for users who don't wish to use
 the IO::File methods.
 
-=over
+=over 4
 
 In the following C<DSH> refers to a data set handle such as returned
 by the C<mvsopen> routine.  For OS data sets C<NAME> refers to either a
 double slashed name such as C<//BETTY.BAM>, or members such as
 C<//BETTY.BAM(BAM)>; or to dd names such as C<dd:WILMA.PEBBLES>.
 
+=item dynalloc HASHREF
+
+Dynamically allocates a data set via the C RTL C<dynalloc()> routine.
+Returns a true value on success, undef on failure.
+
+You may wish to refer to your system's F</usr/include/dynit.h> header
+file for information on the __dyn_t struct typedef as well as
+constants used by the C version of dynalloc().  You might also be 
+interested in symbolic constant names as can be found in dynit.ph 
+after running h2ph on dynit.h (see the INSTALL document for perl).
+
+The hashref to be passed to dynalloc may contain keys with names 
+derived from the __dyn_t member names with the two leading 
+underscores removed.  For example:
+
+    my $hashref = {("ddname" -=> "MYDD", "dsname" => "FRED.DSN", ... )};
+
+While most of the hash values can be character strings (SvPV below)
+some of the hash values must be integers or chars ((cast)SvIV below):
+
+    Perl      C __dyn_t      Perl -> C
+    hash key  member         value type
+    ddname     __ddname =    SvPV(hval,len);
+    dsname     __dsname =    SvPV(hval,len);
+    sysout     __sysout =    (char)SvIV(hval);
+    sysoutname __sysoutname = SvPV(hval,len);
+    member     __member =    SvPV(hval,len);
+    status     __status =    (char)SvIV(hval);
+    normdisp   __normdisp =  (char)SvIV(hval);
+    conddisp   __conddisp =  (char)SvIV(hval);
+    unit       __unit =      SvPV(hval,len);
+    volser     __volser =    SvPV(hval,len);
+    dsorg      __dsorg =     (short)SvIV(hval);
+    alcunit    __alcunit =   (char)SvIV(hval);
+    primary    __primary =   SvIV(hval);
+    secondary  __secondary = SvIV(hval);
+    dirblk     __dirblk =    SvIV(hval);
+    avgblk     __avgblk =    SvIV(hval);
+    recfm      __recfm  =    (short)SvIV(hval);
+    blksize    __blksize =   (short)SvIV(hval);
+    lrecl      __lrecl =     (unsigned short)SvIV(hval);
+    volrefds   __volrefds =  SvPV(hval,len);
+    dcbrefds   __dcbrefds =  SvPV(hval,len);
+    dcbrefdd   __dcbrefdd =  SvPV(hval,len);
+    misc_flags __misc_flags = (unsigned char)SvIV(hval);
+    password   __password =  SvPV(hval,len);
+    miscitems  __miscitems = (char **)SvPV(hval,len);
+    infocode   __infocode =  (short)SvIV(hval);
+    errcode    __errcode =   (short)SvIV(hval);
+    storclass  __storclass = SvPV(hval,len);
+    mgntclass  __mgntclass = SvPV(hval,len);
+    dataclass  __dataclass = SvPV(hval,len);
+    recorg     __recorg =    (char)SvIV(hval);
+    keyoffset  __keyoffset = (short)SvIV(hval);
+    keylength  __keylength = (short)SvIV(hval);
+    refdd      __refdd =     SvPV(hval,len);
+    like       __like =      SvPV(hval,len);
+    dsntype    __dsntype =   (char)SvIV(hval);
+    pathname   __pathname =  SvPV(hval,len);
+    pathopts   __pathopts =  SvIV(hval);
+    pathmode   __pathmode =  SvIV(hval);
+    pathndisp  __pathndisp = (char)SvIV(hval);
+    pathcdisp  __pathcdisp = (char)SvIV(hval);
+
+See also the B<C/C++ Run-Time Library Reference> for information on 
+C<dynalloc()> and C<__dyn_t>.  See also C<svc99>.
+
+=item dynfree HASHREF
+
+Deallocates a data set via the C RTL C<dynfree()> routine.  Returns a 
+true value on success, undef on failure.  For information
+on the form of the HASHREF see C<dynalloc>.  Note that the only __dyn_t
+struct members that are used by the underlying dynfree() rotuine are:
+
+    ddname
+    dsname
+    member
+    pathname
+    normdisp
+    pathndisp
+    miscitems
+
+See also C<svc99>.
+
 =item flush EXPR
 
 This function causes the contents of stdio buffers for the specified
@@ -234,7 +322,7 @@ Additional C<MODE> keyword parameters can be passed from:
     qw(acc= blksize= byteseek lrecl= recfm= type= asis password= noseek)
 
 (See the B<C/C++ MVS Programming Guide> and the 
-B<C/C++ MVS Library Reference> descriptions of C<fopen()> for detailed 
+B<C/C++ Run-Time Library Reference> descriptions of C<fopen()> for detailed 
 information on C<NAME> and C<MODE> arguments.)  If successful, C<mvsopen> 
 returns a data set handle; if an error occurs, it returns C<undef>.
 
@@ -265,7 +353,18 @@ For example:
 
 =item pds_mem NAME
 
-Returns a list of members (excluding ALIASes) for the named PDS directory.  
+=item pds_mem NAME, FLAG
+
+Returns a list of members for the named PDS directory.  Alias names 
+may be returned depending on the value of the optional 
+FLAG argument:
+
+    FLAG   pds_mem() returns
+           member names (if any) - this is the default
+    0      member names (if any)
+    1      alias names only (if there are any)
+    2      member and alias names (if any) 
+
 A list with a single C<undef> element is returned for PDS directories that 
 have no members as well as for data set names that are not partitioned (in
 the latter case a warning may appear on STDERR depending on how 
@@ -274,9 +373,15 @@ For example:
 
     use OS390::Stdio qw(pds_mem);
     my @member_list = pds_mem("//'SLATE.PDS'");
+    print " Members that are not aliases are:\n";
     foreach my $mem (@member_list) {
         print "SLATE.PDS($mem)\n";
     }
+    print " Aliases are:\n";
+    my @alias_list = pds_mem("//'SLATE.PDS'",1);
+    foreach my $alias (@alias_list) {
+        print "SLATE.PDS($alias)\n";
+    }
 
 =item remove NAME
 
@@ -335,9 +440,9 @@ See also C<mvsopen>, C<vsamdelrec>, and C<vsamlocate>.
 
 =back
 
-The following functions are experimental:
-V 0.001: Some are not currently working and either produce fatal 
-errors or simply do not work as intended.
+The following functions are experimental.  Some are not currently 
+working and either produce fatal errors or simply do not work as 
+intended.
 
 =over
 
@@ -353,24 +458,10 @@ Until this is working properly you can from perl code things such as:
 
     @listcat = `tso listcat`;
 
-=item dynalloc 
-
-Dynamically allocates a data set via the C RTL C<dynalloc()> routine.
-See also C<svc99>.
-
-V 0.003: This routine is not yet implemented and causes a fatal error.
-
-=item dynfree 
-
-Deallocates a data set via the C RTL C<dynfree()> routine.
-See also C<svc99>.
-
-V 0.003: This routine is not yet implemented and causes a fatal error.
-
 =item svc99
 
 This function provides access to the SVC 99 system service via a
-C RTL C<svc99()> call.  See the B<C/C++ MVS Library Reference> for
+C RTL C<svc99()> call.  See the B<C/C++ Run-Time Library Reference> for
 information on C<svc99()>.
 
 V 0.003: This routine is not yet implemented and causes a fatal error.
@@ -384,9 +475,77 @@ V 0.003: This routine is not yet implemented and causes a fatal error.
 
 =back
 
+=head1 DIAGNOSTICS
+
+The following messages may be seen when programming with this
+module:
+
+=over 4
+
+=item Data set %s [filename %s] does not appear to be a PDS directory.
+
+Seen during a call to pds_mem() if the named data set does not have the
+__dsorgPDSdir organization and if the module was compiled without -DNO_WARN_IF_NOT_PDS.
+Try calling pds_mem() with the name of a PDS or re-install this module
+being sure to specify -DNO_WARN_IF_NOT_PDS duruing the build process.
+
+=item FREAD: Failed in %s, line %d Expected to read %d bytes but read %d bytes
+
+An error occurred while attempting to fread() a PDS.
+
+=item EFREAD
+
+An error occurred while attempting to fread() a PDS.
+
+=item malloc failed for %d bytes
+
+An error occurred while attempting to malloc() space for a PDS member name.
+
+=item ENONMEM
+
+An error occurred while attempting to malloc() space for a PDS member name.
+
+=item too many args
+
+Seen if an attempt to call pds_mem() with more than 2 arguments
+is made.  pds_mem() ought to be called with a PDS name and an
+optional integer.  Try reducing the list of items passed to
+pds_mem() to one or two.
+
+=item alias flag must be an integer
+
+Seen if the optional second argument passed to pds_mem() is
+not an integer.  Try using an integer expression that evaluates
+to 0 or 1 or 2 instead.
+
+=item dynalloc() requires a hash reference
+
+=item dynalloc() called with undefined value.
+
+=item dynalloc() failed with error code %hX, info code %hX
+
+=item dynalloc() unable to initialize struct __dyn_t
+
+=item dynfree() requires a hash reference
+
+=item dynfree() called with undefined value.
+
+=item dynfree() failed with error code %hX, info code %hX
+
+=item dynfree() unable to initialize struct __dyn_t
+
+=item %s not yet implemented
+
+Seen if an attempt to call an unfinished sub routine is made.
+dsname_level(), svc99(), and vol_ser() are not yet implemented.
+
+=back
+
 =head1 REVISION
 
-This document was last revised on 14-Apr-2001, for Perl 5.6.1.
+This document was last revised on 18-May-2001, for Perl 5.6.1.
+
+14-Apr-2001, VERSION 0.004 for Perl 5.6.1.
 
 13-Apr-1999, VERSION 0.003 for Perl 5.005_03.
 
